{# Import the theme's layout. #}
{% extends "!layout.html" %}

{# Add some extra stuff before and use existing with 'super()' call. #}
{%- block extrahead %}
{{ super() }}
<!-- Font Awesome -->
<link href="https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css" rel="stylesheet">
<!-- favicons via realfavicongenerator.net -->
<link rel="apple-touch-icon" sizes="180x180" href="_static/favicons/apple-touch-icon.png">
<link rel="icon" type="image/png" sizes="32x32" href="_static/favicons/favicon-32x32.png">
<link rel="icon" type="image/png" sizes="16x16" href="_static/favicons/favicon-16x16.png">
<link rel="manifest" href="_static/favicons/site.webmanifest">
<link rel="mask-icon" href="_static/favicons/safari-pinned-tab.svg" color="#107d79">
<link rel="shortcut icon" href="_static/favicons/favicon.ico">
<meta name="apple-mobile-web-app-title" content="ProbNum">
<meta name="application-name" content="ProbNum">
<meta name="msapplication-TileColor" content="#00aba9">
<meta name="msapplication-config" content="_static/favicons/browserconfig.xml">
<meta name="theme-color" content="#ffffff">
<!-- end favicons -->
{% endblock %}